#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# iMembrane : Annotation projection module
# using CGDB simulation data
#
# Author: Sebastian Kelm
# Created: 28/01/2008
#
# Revisions:
# 29/01/2008    Sebastian Kelm    Initial version
# 16/04/2008    Sebastian Kelm    Conversion to object-oriented code
# 30/04/2008    Sebastian Kelm    Merger with ALI file handling
# 01/02/2010    Sebastian Kelm    Split from ALI. Now includes rewritten structre handling (no longer needs PyMOL).
# 06/05/2010    Sebastian Kelm    Complete rewrite of annotation projection code, which is now separate from CGDB interpretation
#





import sys
import os
import re
import shutil
from array import array

from prosci.common import *
from prosci.util.pdb import Pdb
from prosci.util.ali import Ali
from prosci.util.fasta import Fasta
from prosci.util.gaps import gappify, deGappify, isGap, length_ungapped
from prosci.util.seq import pid
from prosci.util.tmalign import superimpose, PoorSuperpositionError
from prosci.util.joy import annotate_single_chain


#regex_tm = re.compile(r"^TM-score=(\d*\.\d*)$", re.MULTILINE)
regex_imemid = re.compile(r"^([a-zA-Z0-9]{4})\.(\d+)\.(.)\.(\d+)(?:\.([0-9x]+))?$")

def splitid(imemid):
  m = regex_imemid.match(imemid)
  if not m:
    raise ValueError("iMembrane id has the wrong format: %s"%imemid)
  values = list(m.groups())
  if not values[4]:
    values[4] = "0"
  return values

def joinid(values):
  assert len(values) in (4,5)
  if len(values) == 4 or not values[4] or values[4]=="0":
    return "%s.%s.%s.%s"%tuple(values[0:4])
  return "%s.%s.%s.%s.%s"%tuple(values)

def imemid2pdb(imemid):
  fields = splitid(imemid)
  return fields[0]+fields[2]


def letter2cssClass(c):
  if c in "NHT":
    return c
  if isGap(c):
    return "gap"
  return "unknown"


def annot2html(seq, prefix="membrane_"):
  if not seq:
    return ""
  
  html = '<span class="%s%s">%s' % (prefix, letter2cssClass(seq[0]), seq[0])
  for c in seq[1:]:
    if c == html[-1]:
      html += c
    else:
      html += '</span><span class="%s%s">%s' % (prefix, letter2cssClass(c), c)
  html += '</span>'
  
  return html


def makeHTMLrow(title, desc, htmlseq):
    title=title.replace('>', '&gt;').replace('<', '&lt;')
    desc=desc.replace('>', '&gt;').replace('<', '&lt;')
    return '<tr class="seqalirow"><td class="seqtitle">%s</td><td class="seqdesc">%s</td><td class="seq">%s</td></tr>\n' % (title, desc, htmlseq)


class Projector:
    #class_names = ("membrane contact", "membrane layer")
    class_names = ("membrane contact", "membrane layer", "accessibility", "membrane consensus")
    
    
    def get_tem(self, dbid):
      return Ali(file("%s/%s/annotation.tem"%(self.dbdir, dbid)))

    def get_structure(self, dbid):
      return Pdb(dbid, file("%s/%s/contact.atm"%(self.dbdir, dbid)))

    def get_planes(self, dbid):
      f = file("%s/%s/layer.txt"%(self.dbdir, dbid))
      txt = f.read()
      f.close()
      return [float(z) for z in txt.strip().split()]

    
    def __init__(self, dbdir, dbid, inputid, ali_file_name, struc_file_name, skipalign=False, outdir=None, insubdir=None, run_joy=False):
      
      if None == dbdir:
        if insubdir and insubdir != os.path.basename(insubdir):
            dbdir=os.path.dirname(insubdir)
            insubdir=os.path.basename(insubdir)
        else:
            raise ArgumentError("Must provide a dbdir")
      
      if not os.path.exists(dbdir):
        raise ArgumentError("dbdir is not a valid path : %s"%(dbdir))
      
      if os.path.exists("%s/entries"%dbdir):
        dbdir = "%s/entries"%dbdir
      
      if dbid:
        if not insubdir:
          insubdir = dbid
        
        if not os.path.exists("%s/%s"%(dbdir, insubdir)):
          raise ArgumentError("insubdir is not a valid path : %s/%s"%(dbdir, insubdir))
      
      if ali_file_name is None and struc_file_name is None:
        raise ArgumentError("Must provide a sequence or structure filename (or both). None given")
    
      
      self.dbdir   = dbdir
      self.dbid    = dbid
      self.insubdir  = insubdir
      
      self.outdir  = "."
      if None!=outdir:
        self.outdir = outdir
      
      
      self.struc   = None
      self.dbstruc = None
      self.planes  = None
      self.inputid = inputid
      self.scores  = {}
      self.queryannot = None
      self.run_joy = run_joy
      
      
      if struc_file_name is not None:
        if isinstance(struc_file_name, Pdb):
          self.struc = struc_file_name.deep_copy()
        else:
          self.struc = Pdb(struc_file_name)
        
        if None == self.inputid:
          self.inputid = self.struc.code
        
        if self.dbid:
          self.dbstruc = self.get_structure(insubdir)
          self.planes  = self.get_planes(insubdir)

      self.ali     = None
      
      if not self.dbid:
        # This is a dummy Projector object.
        
        if None == ali_file_name:
          self.ali = Ali(">%s\nstructure\n%s*\n"%(self.inputid, self.struc.get_seq()))
        else:
          if isinstance(ali_file_name, Ali):
            self.ali = Ali(ali_file_name)
          else:
            self.ali = Ali(file(ali_file_name))
          if None == self.inputid:
            if not len(self.ali)==1:
              raise ArgumentError("Must provide inputid if providing more than one sequence and no structure.")
            lst = list(self.ali)
            self.inputid = lst[0].getCode()
        
        if self.struc is not None:
          if self.struc.chaincount() != 1:
            raise ArgumentError("More than one chain given. Please edit the input structure to contain a single chain only.")
          if deGappify(self.ali[self.inputid].getMasterEntry().get_seq()) != self.struc.get_seq():
            raise ArgumentError("Amino acids in structure are different from alignment/sequence file")
        
        return
      
      
      if self.inputid == self.dbid:
        raise ArgumentError("inputid is equal to the dbid (cannot have duplicate sequence titles in an ALI/TEM file) : %s"%self.inputid)
      
      
      dbannot = self.get_tem(insubdir)
      
      if None == ali_file_name:
        self.ali = Ali(">%s\nstructure\n%s*\n"%(self.inputid, self.struc.get_seq()))
        self.ali.add(dbannot, merge_duplicates=True)
      else:
        if isinstance(ali_file_name, Ali):
          self.ali = Ali(ali_file_name)
        else:
          self.ali = Ali(file(ali_file_name))
        if self.inputid is None:
          if not ((self.ali.has_entry(self.dbid) and len(self.ali)<=2) or len(self.ali)==1):
            raise ArgumentError("Must provide inputid if providing more than one sequence and no structure.")
          lst = list(self.ali)
          if lst[0].getCode() == self.dbid:
            self.inputid = lst[1].getCode()
          else:
            self.inputid = lst[0].getCode()
        self.ali.add(dbannot, merge_duplicates=True)
      
      
      if self.struc is not None:
        if deGappify(self.ali[self.inputid].getMasterEntry().get_seq()) != self.struc.get_seq():
          sys.stderr.write(deGappify(self.ali[self.inputid].getMasterEntry().get_seq()))
          sys.stderr.write("\n")
          sys.stderr.write(self.struc.get_seq())
          sys.stderr.write("\n")
          raise ValueError("Input structure and sequence not consistent: %s, %s"%(self.inputid, self.dbid))
      if self.dbstruc is not None:
        if deGappify(self.ali[self.dbid].getMasterEntry().get_seq()) != self.dbstruc.get_seq():
          sys.stderr.write(deGappify(self.ali[self.dbid].getMasterEntry().get_seq()))
          sys.stderr.write("\n")
          sys.stderr.write(self.dbstruc.get_seq())
          sys.stderr.write("\n")
          raise ValueError("Database structure and sequence not consistent: %s, %s"%(self.inputid, self.dbid))
      
      
      # Align sequences / structures
      #
      isAligned = len(self.ali[self.dbid].getMasterEntry().get_seq()) == len(self.ali[self.inputid].getMasterEntry().get_seq())
      if not skipalign or not isAligned:
        if struc_file_name is None:
          aligned = Fasta("alignment", ">seq1\n%s\n>seq2\n%s\n"%(deGappify(self.ali[self.inputid].getMasterEntry().get_seq()), deGappify(self.ali[self.dbid].getMasterEntry().get_seq())))
          aligned.align()
          self.ali[self.inputid].getMasterEntry().seq = aligned[0].get_seq()
          self.ali[self.dbid].getMasterEntry().seq    = aligned[1].get_seq()

          self.scores["TM-score"] = -1.0
          self.scores["rmsd"] = -1.0
        else:
          seq1, seq2, alignment_info = superimpose(self.struc, self.dbstruc, modify_structures=True, normalise_by_first=True)
          self.ali[self.inputid].getMasterEntry().seq = seq1
          self.ali[self.dbid].getMasterEntry().seq    = seq2
          
          self.scores["TM-score"] = alignment_info["TM-score"]
          self.scores["rmsd"] = alignment_info["RMSD"]
      
      _pid, _pcov, _ncov, _l1, _l2 = pid(self.ali[self.inputid].getMasterEntry().get_seq(), self.ali[self.dbid].getMasterEntry().get_seq(), "first")
      self.scores["pid"]  = _pid # percentage sequence identity
      self.scores["pcov"] = _pcov # percentage sequence coverage
      
      # Make sure each entrygroup is aligned to its master sequence/structure
      self.ali.gappify()
      
      self.ali = Ali([self.ali[self.inputid], self.ali[self.dbid]])
      
      if self.struc:
          s1 = deGappify(self.ali[self.inputid][0].seq)
          s2 = self.struc.get_seq()
          try:
              assert len(s1) == len(s2)
          except:
              print repr(s1)
              print repr(s2)
              print type(s1), type(s2)
              for x, y in zip(s1, s2):
                if x != y: print x, y
              raise
    
    
    
    def annotate(self, replace=True):
        self.make_all_basic_annotation(replace=replace)

        if self.struc is not None:
            self.extend_annotation()
        
        if not "pannot" in self.scores:
            # Calculate percentage of sequence annotated.
            # We pretend that the entire sequence is the TM domain,
            # even if we know some bits are soluble. This should balance out the
            # fact that we will have large unaligned regions that may correspond
            # to soluble domains. These will drag down the scores, but we
            # have no way of being sure that a large alignment gap is truly a
            # soluble domain...
            #
            contact = self.ali[self.inputid]["membrane contact"].get_seq()
            length = length_ungapped(contact)
            nannot = length - contact.count("?")
            self.scores["pannot"] = float(nannot) / length
    
    
    
    def write_output(self, outdir, subdir=None, link_parent=False, copy_parent=False, write_fasta=False):
      
      self.annotate_query()
      
      entrycode = self.inputid
      if subdir is None:
        subdir = entrycode
      
      entrydir = os.path.join(outdir, subdir)
      if not os.path.exists(entrydir):
          os.mkdir(entrydir)
      
      f = open(os.path.join(entrydir, "annotation.tem"), 'w')
      f.write(str(self.ali))
      f.close()
      
      if self.queryannot:
        f = open(os.path.join(entrydir, "query_annotation.tem"), 'w')
        f.write(str(self.queryannot))
        f.close()
        
        f  = open(os.path.join(entrydir, "query_annotation.inc.html"), 'w')
        for e in self.queryannot[0]:
          if e.isMaster():
            f.write(makeHTMLrow(e.getCode(), e.getDesc(), e.get_seq()))
          elif e.getType() in Projector.class_names:
            if e.getType() not in ("membrane layer", "membrane contact"):
              continue
            f.write(makeHTMLrow(e.getCode(), e.getDesc(), annot2html(e.get_seq())))
        f.close()
      
      if write_fasta:
        f = open(os.path.join(entrydir, "sequence.fasta"), 'w')
        f.write(self.ali[entrycode].getMasterEntry().toFastaString(remove_gaps=True))
        f.close()
      
      
      f  = open(os.path.join(entrydir, "annotation.inc.html"), 'w')
      for eg in self.ali:
        for e in eg:
          if e.isMaster():
            f.write(makeHTMLrow(e.getCode(), e.getDesc(), e.get_seq()))
          elif e.getType() in Projector.class_names:
            if e.getType() not in ("membrane layer", "membrane contact"):
              continue
            f.write(makeHTMLrow(e.getCode(), e.getDesc(), annot2html(e.get_seq())))
      f.close()
      
      
      f = open(os.path.join(entrydir, "scores.table"), 'w')
      f.write("inputid\t%s\n"%(self.inputid))
      f.write("dbid\t%s\n"%(self.dbid))
      for sc in sorted(self.scores):
        f.write("%s\t%g\n"%(sc, self.scores[sc]))
      f.close()
      
      if self.struc is not None:
        f = open(os.path.join(entrydir, "layer.txt"), 'w')
        f.write(join(" ", self.planes))
        f.write("\n")
        f.close()
        
        for cl in Projector.class_names:
          self.annotate_structure(cl)
          
          shortclass = cl.split()[-1]
          
          f=open(os.path.join(entrydir, shortclass+".atm"), 'w')
          f.write(str(self.struc))
          f.close()
          
          if cl=="membrane contact":
            os.system('ln -sf "%s.atm" "%s.atm"'%(shortclass, os.path.join(entrydir, entrycode)))
      
      os.system('chmod a+r "%s/"*.*'%(entrydir))
      if copy_parent:
        newname = os.path.join(entrydir, "parent")
        source = os.path.join(self.dbdir, self.insubdir)
        #if not os.path.isabs(self.dbdir):
        #  source = os.path.relpath(source, entrydir)
        shutil.copytree(source, newname)
      elif link_parent:
        newname = os.path.join(entrydir, "parent")
        source = os.path.join(self.dbdir, self.insubdir)
        if not os.path.isabs(self.dbdir):
          source = os.path.relpath(source, entrydir)
        os.system('ln -sfn "%s" "%s"'%(source, newname))
      
      os.system('chmod a+rx "%s"' % (entrydir))
    
    
    
    def make_all_basic_annotation(self, replace=True):
        class_strings = [self.ali[self.dbid][cl].get_seq() for cl in Projector.class_names]
        
        inputseq = self.ali[self.inputid].getMasterEntry().get_seq()
        
        for i,s in enumerate(class_strings):
          assert len(inputseq) == len(s)
          myannot=array('c')
          for j,c in enumerate(s):
            if not isGap(inputseq[j]):
              if isGap(c):
                myannot.append('?')
              else:
                myannot.append(c)
            else:
              myannot.append('-')
          
          cl = Projector.class_names[i]
          self.ali.add_entry(Ali.Entry(self.inputid, cl, myannot.tostring()), replace_duplicates=replace)
        
        self.queryannot = None
    
    
    
    def z2class(self, zcoord):
      if   zcoord > self.planes[0] or zcoord < self.planes[3]:
          # Residue is OUTSIDE the membrane, in an aqueous environment
          return 'N'
      elif zcoord > self.planes[1] or zcoord < self.planes[2]:
          # Residue is in the HEAD group layer of the membrane
          return 'H'
      else:
          # Residue is in the TAIL group layer of the membrane
          return 'T'
    
    
    
    def extend_annotation(self):
        zcoords = [x.z for x in self.struc.get_CA()]
        
        entrygroup = self.ali[self.inputid]
        masterseq  = entrygroup.getMasterEntry().get_seq()
        
        output = array('c')
        for z in zcoords:
            output.append(self.z2class(z))
        
        try:
          layers = gappify(masterseq, output)
        except ValueError:
          sys.stderr.write("Entrycode:\n%s\n\nDB-entry:\n%s\n\nInput:\n%s\n" % (self.inputid, masterseq, output.tostring()))
          raise
        
        entrygroup["membrane layer"].seq = layers
        
        contact = array('c', entrygroup["membrane contact"].seq)
        unannotated=0
        tm_domain_size=0
        for i,v in enumerate(contact):
            if layers[i] in "HT":
                tm_domain_size+=1
            if v == '?':
                if layers[i] == 'N':
                   contact[i] = layers[i]
                else:
                    unannotated += 1
        
        # This definition of pannot (%annotated) is different from the one
        # for sequence-based annotation. Here we define the transmembrane domain
        # as everything in the H and T layers and ignore all soluble bits.
        #
        if tm_domain_size > 0:
          self.scores["pannot"] = 1.0 - float(unannotated)/tm_domain_size
        else:
          self.scores["pannot"] = 1.0
        
        entrygroup["membrane contact"].seq = contact.tostring()
        
        self.queryannot = None
    
    
    
    def annotate_structure(self, annot_type):
        assert annot_type in Projector.class_names
        
        class_string = deGappify(self.ali[self.inputid][annot_type].get_seq())
        
        try:
          assert len(class_string) == self.struc.rescount()
        except:
          print len(class_string), self.struc.rescount()
          print class_string
          print self.struc.get_seq()
          raise
        
        for i,res in enumerate(self.struc.xresidues()):
          c = class_string[i]
          
          b = 0.0
          if   'N'==c:
              b = 1.0
          elif 'H'==c:
              b = 100.0
          elif 'T'==c:
              b = 200.0
          
          for atm in res:
            atm.b = b
    
    
    
    def annotate_query(self):
        if self.queryannot:
          return
        
        annot = Ali([self.ali[self.inputid]])
        annot.remove_gaps()
        
        if self.run_joy and self.struc:
          joyannot = annotate_single_chain(self.struc, self.inputid)
          annot.add(joyannot, merge_duplicates=True, replace_duplicates=True)
          annot.remove_gaps()
        
        self.queryannot = annot
