#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
import tempfile
import subprocess
import shlex
import shutil
import traceback

from prosci.common import *
from prosci.util.pdb import Pdb
from prosci.util.pdb3d import averageCoordinates
from prosci.util.residue import ResidueList, is_residue_consecutive, is_residue_numbering_consecutive
from prosci.util.ali import Ali
from prosci.loops.multifread import MultiFread

#class DependencyError(RuntimeError):
#  pass
#class ExecutableDependencyError(DependencyError):
#  pass
#class DatabaseDependencyError(DependencyError):
#  pass

def getFreadPath():
  freadpath = os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])), "multifread")
  
  if not freadpath or not os.path.exists(freadpath):
    p = subprocess.Popen("which multifread 2>/dev/null", shell=True, stdout=subprocess.PIPE)
    freadpath, errors = p.communicate()
    freadpath = freadpath.strip()
    
    if not freadpath or not os.path.exists(freadpath):
      return None
    
    freadpath = os.path.abspath(freadpath)
  
  return freadpath


def getFreadDatabaseDir(dbtype="mem", version="5", uniqdb=None):
  path = getFreadPath()
  if not path:
    raise ExecutableDependencyError("FREAD not found. Make sure the FREAD directory is in your PATH.")
  path = os.path.dirname(os.path.abspath(os.path.dirname(path)))
  
  if uniqdb is None:
    uniqdb = set([])
  
  db_level1 = []
  for i, db1 in enumerate(dbtype.split(":")):
    db_level2 = []
    for j, db2 in enumerate(db1.split("|")):
      db2_path = None
      for candidate_path in (db2, "db_"+db2, "db_"+db2+version):
        
        if os.path.isdir(candidate_path):
          db2_path = candidate_path
          break
        if os.path.isfile(candidate_path):
          try:
            db2_path = getFreadDatabaseDir(read_file(candidate_path).strip(), version, uniqdb)
            break
          except DatabaseDependencyError:
            pass
        
        candidate_path = os.path.join(path, candidate_path)
        
        if os.path.isdir(candidate_path):
          db2_path = candidate_path
          break
        if os.path.isfile(candidate_path):
          try:
            db2_path = getFreadDatabaseDir(read_file(candidate_path).strip(), version, uniqdb)
            break
          except DatabaseDependencyError:
            pass
        
      if db2_path:
        if db2_path not in uniqdb:
          db_level2.append(db2_path)
          uniqdb.add(db2_path)
      else:
        sys.stderr.write("NOTICE: FREAD database of type '%s' is not in the expected location '%s'. Skipping it.\n" % (db2, path))
    if db_level2:
      db_level1.append("|".join(db_level2))
  if db_level1:
    path = ":".join(db_level1)
  else:
    raise DatabaseDependencyError("FREAD is installed but none of the specified databases ('%s') were found" % (dbtype))
  
  return path


MAXLOOPLEN = 26
ANCHORLEN  = 2
SHORTLOOPLEN = 3


def startsWithAny(query, prefixes):
    for x in prefixes:
      if query.startswith(x):
        return True
    return False



###############################################################################



def modelAllGaps(core_struc_filename, target_seq, core_target_residue_indeces, outfilename, decoy_dir_prefix="decoys", loop_log_filename=None, score_cutoff=25, max_extension=1000000, core_struc=None, dbtype="mem", summary_file_prefix="summary", first_decoys=0, ranking=None, minb=0, exclude_pdbs=None, anchor_rmsd=1.0, contact_identity=-1.0, pool_size=1, **kwargs):
    
    if loop_log_filename is None:
      loop_log_filename = "-"
    
    # Formatting the target sequence, such that residues with corresponding co-ordinates are upper case and missing residues are lower case.
    #
    
    indexset = []
    #if core_struc is not None:
    #  for start, end in core_struc.xresidue_boundaries():
    #    indexset.append(core_struc[start].ires)
    #indexset = set(indexset)
    
    seq = ""
    j=0
    for i in core_target_residue_indeces:
      prefix = target_seq[j:i]
      if prefix:
        if indexset and (j+1 in indexset):
          seq += prefix.upper()
        else:
          seq += prefix.lower()
      seq += target_seq[i].upper()
      j=i+1
    seq += target_seq[j:].lower()
    
    
    filter = None
    if exclude_pdbs:
      exclude_pdbs = tuple([x[:4].upper() for x in exclude_pdbs])
      filter = lambda decoy: decoy.struc[:4].upper() not in exclude_pdbs
    
    try:
      mf = MultiFread(core_struc_filename, seq)
      #mf.set_subst_tables(opts.esst)
      models = mf.model_structure(db=getFreadDatabaseDir(dbtype),
                         score = score_cutoff,
                         open_rmsd = anchor_rmsd,
                         #closed_rmsd = 0.3,
                         #vdw_factor = 0.7,
                         #max_melding_rmsd = 1.0,
                         #nostrucdir = False, 
                         nofileout = True,
                         #noclose = False, 
                         #nomeld = False,
                         #extension_min = 0,
                         extension_max = max_extension,
                         summary = summary_file_prefix,
                         strucdir = decoy_dir_prefix,
                         outfile = outfilename,
                         messages = loop_log_filename,
                         max_decoys = 10,
                         first_decoys = first_decoys,
                         top_models = 1,
                         pool_size = pool_size,
                         verbose = True,
                         debug = (pool_size == 1),
                         f_rank_decoy = ranking,
                         f_filter = filter,
                         min_b_factor = minb,
                         calculate_contacts = contact_identity >= 0,
                         contact_identity = contact_identity,
                         **kwargs)
    #except DatabaseDependencyError:
    #  raise
    except:
      traceback.print_exc()
      # This will actually mean MEDELLER won't crash, even if FREAD did...
      shutil.copy(core_struc_filename, outfilename)
      return None
    
    
    #for line in out.splitlines():
    #  if line.startswith("WARNING:") or line.startswith("ERROR:"):
    #    print line
    
    if models:
      f = open(outfilename, "w")
      f.write(str(models[0]))
      f.close()
      return True
    else:
      if core_struc_filename != outfilename:
        shutil.copy(core_struc_filename, outfilename)
    
    return False


def countMaxModellableResidues(coords, max_loop_length=None):
  covered = coords.rescount()
  gappy = 0
  
  if max_loop_length is None:
    max_loop_length = MAXLOOPLEN
  
  gl = coords.get_gap_list()
  # Format: gl = [[startres, length, index], ...]
  for g in gl:
    l = g[1]
    if l <= max_loop_length:
      gappy += l
  
  return gappy+covered


def calculateCoreCoverage(coords, max_loop_length=None):
  covered = coords.rescount()
  gappy = 0
  
  if max_loop_length is None:
    max_loop_length = MAXLOOPLEN
  
  gl = coords.get_gap_list()
  # Format: gl = [[startres, length, index], ...]
  for g in gl:
    l = g[1]
    if l <= max_loop_length:
      gappy += l
  
  return float(covered) / float(gappy+covered)




# def mapCore2TargetSeq_alignment(seq, model):
#   a=Ali(">target\nsequence\n%s\n>%s\nsequence\n%s\n" % (seq, model.code, model.get_seq()))
#   a.align()
#   indeces=[]
#   for i, (t, m) in enumerate(zip(a["target"].master.seq, a[model.code].master.seq)):
#     assert not isGap(t)
#     if not isGap(m):
#       assert m.upper() == t.upper(), "%s\n%s\n"%(a["target"].master.seq, a[model.code].master.seq)
#       indeces.append(i)
#   return indeces, a


def mapCore2TargetSeq(seq, model, targetid="target", modelid=None, renumber=False):
  if modelid is None:
    modelid = model.code
  
  if isinstance(model, ResidueList):
    model_residues = model
  else:
    model_residues = ResidueList(model)
  
  modelseq = model_residues.get_seq()
  assert len(modelseq) == len(model_residues)
  
  model_fragments = [ResidueList([model_residues[0]])]
  for i in xrange(1,len(model_residues)):
    if is_residue_consecutive(model_residues[i-1], model_residues[i]) and is_residue_numbering_consecutive(model_residues[i-1], model_residues[i]):
      model_fragments[-1].append(model_residues[i])
    else:
      model_fragments.append(ResidueList([model_residues[i]]))
  
  seq = seq.upper()
  modelseq_aligned = ""
  indeces = []
  i = 0
  for frag in model_fragments:
    fragseq = frag.get_seq()
    j = seq.find(fragseq, i)
    if j < 0:
      raise Exception("Cannot map model sequence to complete sequence:\n%s\n%s\n%s"%(seq, modelseq, str([f.get_seq() for f in model_fragments])))
    indeces.extend(range(j, j+len(fragseq)))
    modelseq_aligned += "-" * (j-i) + fragseq
    i = j + len(fragseq)
    
    if renumber:
      # Renumber residues to show their position in the native sequence
      # (starting with residue 1)
      #
      for k, residue in enumerate(frag):
        for atom in residue:
          atom.ires = 1 + j + k
          atom.inscode = ""
  
  modelseq_aligned += "-" * (len(seq) - len(modelseq_aligned))
  
  a = Ali(">%s\nsequence\n%s\n>%s\nsequence\n%s\n" % (targetid, seq, modelid, modelseq_aligned))
  
  return indeces, a

